/*
 * Project name:
     PIC32 SD Card Bootloader
 * Copyright:
     GNU General Public License
     use this software but let the community be part of its development
 * Revision History:
     2017 02 28 :
       - initial release and main work done by M.Hooper on
         https://forum.mikroe.com/viewtopic.php?f=220&t=69603&p=280504&hilit=bootloader+hooper#p280504
     2017 10 19 :
       - Adopted to FAT32 File version and EasyPicFusion Board by A. Wolf
 * Description:
     This project is a SD Card Bootloader using FAT 32 on
 * Test configuration:
     MCU:             P32MX795F512L
     dev.board:       easypic fusion v7 - 
     Oscillator:      XT-PLL, 80.0000MHz
     Ext. Modules:    None.
     SW:              mikroC PRO for PIC32
 * NOTES:
 
     enable buttons
     - Turn UP Button Press Level switch SW10 for PORTB
     - PULD DOWN switch for PORTB RB0

     enable SD Card
     -  sw13 & 14
 
     enable Buzzer
     - Turn on Buzzer switch SW14.8.

     ATTENTION - READ THIS !
     - in Types.h Bootloader app size in #define USER_APP_RESRT_ADDR 0x9D00A000
       must be adopted if this code is changed
     - for release version compile this code with min Optimization 1 in order to get
       its size as small as possible
       
     - include these lines:
       #define USER_APP_RESRT_ADDR 0x9D00A000
       #define ConfigMem();  \
          OrgAll(USER_APP_RESRT_ADDR); \
          FuncOrg(__BootStartUp,USER_APP_RESRT_ADDR);

       in your application code and call it in main!

       The Bootloader erases all the flash pages that belong to the program flash from the address of 0x9D004000. this means that if you use some section of this flash to keep your application specific data then it will be lost when you use this bootloader to upgrade your firmware. you can adjust the bootloader application to preserve some space that you need.


*/

#include "types.h"
#include "SdCard.h"
#include "Flash.h"
#include "Bootloader.h"

#define SDCARD_NOT_MOUNTED  1
#define SDCARD_MOUNTED      0
#define P_INPUT  1
#define P_OUTPUT 0
#define P_ON  1
#define P_OFF 0
#define P_HIGH  1
#define P_LOW 0
#define ENABLE_  1
#define DISABLE_ 0


sbit TEST_LED_PORT             at PORTB.B1;
sbit TEST_LED                  at LATB1_bit;
sbit TEST_LED_Direction        at TRISB1_bit;

// SD CARD
sbit Mmc_Chip_Select           at LATD12_bit;
sbit Mmc_Chip_Select_Direction at TRISD12_bit;

sbit CARD_DETECT               at PORTD.B13;
sbit CARD_DETECT_Direction     at TRISD13_bit;

sbit USER_BUTTON_DETECT        at PORTB.B0;
sbit USER_BUTTON               at PORTD.B0;
sbit USER_BUTTON_Direction     at TRISB0_bit;  // Card detect


// ----------------- function prototypes --------------------------------------
void configBoard();
void  launchUserApplication(unsigned short isAfterBootload);
unsigned short isUseApplocationExist();
unsigned short isApplicationExist(void);
//------------------------------------------------------------------------------


//------------------------------------------------------------------------------
void launchUserApplication(unsigned short isAfterBootload){
unsigned short i;

   if (isAfterBootload == TRUE){
      for (i=0 ; i<3; i++){
        Delay_ms(500);
        Sound_Play(4000,1000);  // make some noise before launching new user app.
      }
   }

   R30 = USER_APP_RESRT_ADDR;
   asm {
      JR R30
      NOP
   }
}

//---------------------------------------------
unsigned short isApplicationExist() {

   unsigned long FirsAppWord;

   memcpy(&FirsAppWord,(char *)USER_APP_RESRT_ADDR,4);
   if (FirsAppWord == 0xFFFFFFFF){
       return FALSE;
   } else {
       return TRUE;
   }
}


short enableSDcard(){
  
  SPI3CONbits.ON = 1;
  SPI3_Init_Advanced(_SPI_MASTER, _SPI_8_BIT, 80, _SPI_SS_DISABLE, _SPI_DATA_SAMPLE_MIDDLE, _SPI_CLK_IDLE_HIGH, _SPI_ACTIVE_2_IDLE);
  SPI_Set_Active(SPI3_Read, SPI3_Write); // Sets the SPI1 module active
  Delay_ms(10);

  return 1;
}

// 0  = Fat32 init succesfull
// -1 = FAT32_X_Init() ERROR
// -2 = no SD card inserted
short FAT32_X_Init(){
  short err_intern = 0;

  if(CARD_DETECT == SDCARD_NOT_MOUNTED){
      //UART2_Write_Text("No SD Card inserted\x0D");
      return -2;
      }

  enableSDcard();

  err_intern = FAT32_Init();
  return err_intern;
}



enResult SD_initDriver(){
  int err = 0;
 
  err = FAT32_X_Init();
  if(err < 0){
    TEST_LED = 1;
    }
  else{
       //UART2_Write_Text("SD Card init succesfull\x0D");
       }
  
  return eResultOk;
}


//-------------------------------------
void configBoard(){
  int i = 0;
  JTAGEN_bit = 0;   // disable JTAG
  AD1PCFG = 0xFFFF;
  // as input
  TRISA = 0xFFFF;
  TRISB = 0xFFFF;
  TRISC = 0xFFFF;
  TRISD = 0xFFFF;
  TRISE = 0xFFFF;
  TRISF = 0xFFFF;
  TRISG = 0xFFFF;

  // INPUT / OUTPUT Port Setting ---------------------

  TEST_LED_Direction  = P_OUTPUT;
  USER_BUTTON_Direction = P_INPUT;
  CARD_DETECT_Direction = P_INPUT;     // Card detect

  // ---------------------------------------------------------------------------

  TEST_LED = P_OFF;
  for(i = 0; i < 10; i++){
     TEST_LED = ~TEST_LED_PORT;
     Delay_ms(100);
    }
  TEST_LED = P_ON;

  DisableInterrupts();
  
  //UART2_Init(57600);
  //UART2_Write_Text("\x0DuC started\x0D");

  SD_initDriver();
  Sound_Init(&PORTD, 3);
}





void main() {
   enResult eresult;
   unsigned short i;

 
   configBoard(); // config the board

  // make some sound when starting BL App
   for (i = 0; i < 3; i++){
     //Sound_Play(4000,200);
     TEST_LED = ~TEST_LED_PORT;
     Delay_ms(200);
   }

   //check if we need to enter to BL mode
   i = isBootloaderMode();
   if( i == TRUE) {
      initShadowMemory();
      eresult = runBootloader();
      if (eresult == eResultOk) {
         DisableInterrupts();
         launchUserApplication(TRUE);
      }
   }else {
       if (isApplicationExist() == TRUE){
          DisableInterrupts();
          launchUserApplication(FALSE);
       }

      /*
       If App does not exist it will exit the if
         and continue in the endless while.
         that is at the end of the function
       */
   }
   while (TRUE) {
       Sound_Play(4000,200);
       TEST_LED = ~TEST_LED_PORT;
       Delay_ms(500);
   }
}