#include "types.h"

#define PROGRAM_FLSH_SIZE 0x80000    // 512K
#define BOOT_FLSH_SIZE 0x3000        //12K
#define FLASH_ERASE_BLOCK_SIZE 0x1000 //4K
#define PHY_START_PROGRAM_FLASH 0x1D000000
#define PHY_END_PROGRAM_FLSH    0x1D07FFFF
#define PHY_START_BOOT_FLASH    0x1FC00000
#define PHY_END_BOOT_FLASH      0x1FC02FEF

// ------------------- local functions protitypes ------------------------------
unsigned short getPageIndex(unsigned long phyAddress);
unsigned long GetPhyStartPageAddress(unsigned long phyAddress);
unsigned long physicaToKseg1(unsigned long physicalAddr);
void writeShadowToFlash(void);
//----------------------------- module objects ---------------------------------
typedef struct{
   unsigned short blockIndex;
   unsigned long  phyPageStartAddr;
   unsigned short shadowBlock[FLASH_ERASE_BLOCK_SIZE];
} shadowFlashBlock_t;

//----------------------------- module variables -------------------------------
unsigned short m_ProgramFlashBlocksTrace[131]; // from [0-127] for program Flash [128-130] for Boot Flash
shadowFlashBlock_t m_shadowPage;
unsigned short m_oneShot = FALSE;
//------------------------------------------------------------------------------

void initShadowMemoriesTraces(){
   unsigned short i;
   for (i=0; i< 131; i++){
      //first 128 blocks [0,127] are program flash blocks last three[128,130] are for boot flash
      m_ProgramFlashBlocksTrace[i] = 0; // mark block as not deleted
   }
}
//------------------------------------------------------------------------------
void eraseFlashFromUserApplicationStart(){
   unsigned long  pageStartAddr = USER_APP_RESRT_ADDR; // this is virtual address
   unsigned short pageIndex = getPageIndex(KVA_TO_PA(USER_APP_RESRT_ADDR));
   do {
         Flash_Erase_Page(pageStartAddr);
         m_ProgramFlashBlocksTrace[pageIndex] = 1; //mark as erased
         pageStartAddr += 0x1000;
         pageIndex++;
   }while (pageStartAddr < 0x9D07FFFF);
}
   
//------------------------------------------------------------------------------
void initShadowMemory(){
   initShadowMemoriesTraces();
   m_oneShot = FALSE;
   eraseFlashFromUserApplicationStart();
}
//------------------------------------------------------------------------------
/*
   This function is writing the data of te hex line to the apropriate location
   in the shadow block memory.
   there are two situations when the shadow block has to be flushed to the flash memory:
   1. the arriving hex data location is on a different page in the flash than the shadow block.
   2. the hex line data is crossing the border betwen two pages.
   In both cases the current shadow block should be written to the mcu flash and another bloack should be
   read from flash memory and serve as shadow memory.
   Writing shadow block to the mcu flash should be done when the page in mcu is delleted on the first attempt of write.
   Bear in mind that the same block might appear several times in the hex file.
   
*/
void writeDataToShadowOfFlashBlock(hexLine_t *ptrHexObj){
   unsigned long numBytesPageOverflow = 0;
   unsigned short dataBlockIndex;
   unsigned long offsetInPage;
   unsigned long *ulPtr;
  
   dataBlockIndex = getPageIndex(ptrHexObj->realAddess);
  
   if (m_oneShot == FALSE){
      m_shadowPage.phyPageStartAddr =  GetPhyStartPageAddress(ptrHexObj->realAddess);
      ulPtr = physicaToKseg1(m_shadowPage.phyPageStartAddr);
      memcpy(m_shadowPage.shadowBlock,ulPtr, FLASH_ERASE_BLOCK_SIZE);  // read from flash
      m_shadowPage.blockIndex = dataBlockIndex;
      m_oneShot = TRUE;
   }

   if(m_shadowPage.blockIndex != dataBlockIndex){ // new arriving data is in another flash page than the shadow
      // erase flash page before writing ( if not already erased)
//      if (m_ProgramFlashBlocksTrace[m_shadowPage.blockIndex] == 0){ // page has not been erased
//         virtualAddrOfPageStart = physicaToKseg1(m_shadowPage.phyPageStartAddr);
//         Flash_Erase_Page(virtualAddrOfPageStart); // erase before write
//         m_ProgramFlashBlocksTrace[m_shadowPage.blockIndex] = 1; // mark it as 'has been erased'
//      }
      // write the shadow memory to the flash (eight rows of 512 bytes = 4K)
      writeShadowToFlash();

      // read the new page to the shadow memory and update the object attributes
      m_shadowPage.phyPageStartAddr =  GetPhyStartPageAddress(ptrHexObj->realAddess); 
      ulPtr =  physicaToKseg1(m_shadowPage.phyPageStartAddr);
      memcpy(m_shadowPage.shadowBlock, ulPtr, FLASH_ERASE_BLOCK_SIZE);
      m_shadowPage.blockIndex = dataBlockIndex;
   }
   
   // find the offset in the page and store the data in correct place in shadow memory
   offsetInPage = ptrHexObj->realAddess - m_shadowPage.phyPageStartAddr;
   if ( (offsetInPage + ptrHexObj->dataLength) > FLASH_ERASE_BLOCK_SIZE){    //if the current hex row data spans two adjacent pages
      numBytesPageOverflow =  offsetInPage + ptrHexObj->dataLength - FLASH_ERASE_BLOCK_SIZE;
      // thus fill the shadow untill the end of the current page with data from the hex row
      memcpy(& m_shadowPage.shadowBlock[offsetInPage], ptrHexObj->hexBinData,ptrHexObj->dataLength - numBytesPageOverflow);
      // write the shadow to flash ( erase flash page if needed)
//      if (m_ProgramFlashBlocksTrace[m_shadowPage.blockIndex] == 0){// page has never been erased
//         virtualAddrOfPageStart = physicaToKseg1(m_shadowPage.phyPageStartAddr);
//         Flash_Erase_Page(virtualAddrOfPageStart);                    // erase before write
//         m_ProgramFlashBlocksTrace[m_shadowPage.blockIndex] = 1;       // mark block as already been erased
//      }
      writeShadowToFlash();
      
      // load the adjacent page to the shadow memory
      m_shadowPage.phyPageStartAddr += 0x1000; // adjacent page starts at 4K bytes from the current
      ulPtr =   physicaToKseg1(m_shadowPage.phyPageStartAddr);
      memcpy(m_shadowPage.shadowBlock, ulPtr, FLASH_ERASE_BLOCK_SIZE); // fill shadow from flash ****
      m_shadowPage.blockIndex ++; // the block index is adjacent to the last one

      // now write the remaining bytes from the hex row to the start of the new page
      ulPtr = ptrHexObj->hexBinData + (ptrHexObj->dataLength - numBytesPageOverflow);
      memcpy(m_shadowPage.shadowBlock, ulPtr, numBytesPageOverflow); //****
   } else{
      // all the data in the hex line fits ino the page and not overflows to adjacent page so just write thenm to the shadow.
      memcpy(& m_shadowPage.shadowBlock[offsetInPage], ptrHexObj->hexBinData, ptrHexObj->dataLength);
   }
}
//------------------------------------------------------------------------------
void writeShadowToFlash(void) {
  unsigned long i;
  unsigned long virtualAddrOfPageStart;
  
  virtualAddrOfPageStart =  physicaToKseg1(m_shadowPage.phyPageStartAddr);
  // erase flash page before writing ( if not already erased)
  if (m_ProgramFlashBlocksTrace[m_shadowPage.blockIndex] == 0){ // page has not been erased
     Flash_Erase_Page(virtualAddrOfPageStart); // erase before write
     m_ProgramFlashBlocksTrace[m_shadowPage.blockIndex] = 1; // mark it as 'has been erased'
  }
//  // now write
  for (i=0; i<8; i++){
      Flash_Write_Row(virtualAddrOfPageStart + i*512, & m_shadowPage.shadowBlock[512 *i]);
  }
}
//------------------------------------------------------------------------------
 unsigned long GetPhyStartPageAddress(unsigned long phyAddress){
    //return  ((phyAddress >> 12) << 12);
    return   phyAddress & ~(FLASH_ERASE_BLOCK_SIZE -1);
 }
 //-----------------------------------------------------------------------------
 unsigned short getPageIndex(unsigned long phyAddress){
   if (phyAddress < PHY_START_BOOT_FLASH){
       return  (unsigned short) ((phyAddress - PHY_START_PROGRAM_FLASH)>>12);
    } else {
       return  (unsigned short)(((phyAddress - PHY_START_BOOT_FLASH) >>12)) +128; //boot flash indexes ara 128,129,130
    }
 }
 //-----------------------------------------------------------------------------
 unsigned long physicaToKseg1(unsigned long physicalAddr){
     //KSEG0 Virtual Address = Physical Address | 0xA0000000
     return  physicalAddr | 0x80000000;
 }